@@ -3,6 +3,7 @@
 module Control.Foldl.Text (
     -- * Folding
       fold
+    , foldM
 
     -- * Folds
     , head
@@ -20,6 +21,7 @@ module Control.Foldl.Text (
     , elemIndex
     , findIndex
     , count
+    , lazy
 
     -- * Re-exports
     -- $reexports
@@ -27,95 +29,112 @@ module Control.Foldl.Text (
     , module Data.Text
     ) where
 
-import Control.Foldl (Fold)
-import Control.Foldl.Internal (Maybe'(..), lazy, strict, Either'(..), hush)
-import qualified Control.Foldl as L
+import Control.Foldl (Fold, FoldM)
+import Control.Foldl.Internal (Maybe'(..), strict, Either'(..), hush)
 import Data.Text (Text)
-import qualified Data.Text as T
-import qualified Data.Text.Lazy as Lazy
 import Prelude hiding (
     head, last, null, length, any, all, maximum, minimum, elem, notElem )
 
+import qualified Control.Foldl
+import qualified Control.Foldl.Internal
+import qualified Data.Text
+import qualified Data.Text.Lazy
+
 -- | Apply a strict left 'Fold' to lazy text
-fold :: Fold Text a -> Lazy.Text -> a
-fold (L.Fold step begin done) as = done (Lazy.foldlChunks step begin as)
+fold :: Fold Text a -> Data.Text.Lazy.Text -> a
+fold (Control.Foldl.Fold step begin done) as =
+    done (Data.Text.Lazy.foldlChunks step begin as)
 {-# INLINABLE fold #-}
 
+-- | Apply a strict monadic left 'FoldM' to lazy text
+foldM :: Monad m => FoldM m Text a -> Data.Text.Lazy.Text -> m a
+foldM (Control.Foldl.FoldM step begin done) as = do
+    x <- Data.Text.Lazy.foldlChunks step' begin as
+    done x
+  where
+    step' mx bs = do
+      x <- mx
+      x `seq` step x bs
+{-# INLINABLE foldM #-}
+
 {-| Get the first character of a text stream or return 'Nothing' if the stream
     is empty
 -}
 head :: Fold Text (Maybe Char)
-head = L.Fold step Nothing' lazy
+head = Control.Foldl.Fold step Nothing' Control.Foldl.Internal.lazy
   where
     step mc txt =
-        if T.null txt
+        if Data.Text.null txt
         then mc
         else case mc of
             Just' _  -> mc
-            Nothing' -> Just' (T.head txt)
+            Nothing' -> Just' (Data.Text.head txt)
 {-# INLINABLE head #-}
 
 {-| Get the last character of a text stream or return 'Nothing' if the text
     stream is empty
 -}
 last :: Fold Text (Maybe Char)
-last = L.Fold step Nothing' lazy
+last = Control.Foldl.Fold step Nothing' Control.Foldl.Internal.lazy
   where
     step mc txt =
-        if T.null txt
+        if Data.Text.null txt
         then mc
-        else Just' (T.last txt)
+        else Just' (Data.Text.last txt)
         -- TODO: Use `unsafeLast` when Debian Stable Haskell Platform has it
 {-# INLINABLE last #-}
 
 -- | Returns 'True' if the text stream is empty, 'False' otherwise
 null :: Fold Text Bool
-null = L.Fold step True id
+null = Control.Foldl.Fold step True id
   where
-    step isNull txt = isNull && T.null txt 
+    step isNull txt = isNull && Data.Text.null txt 
 {-# INLINABLE null #-}
 
 -- | Return the length of the text stream in characters
 length :: Num n => Fold Text n
-length = L.Fold (\n txt -> n + fromIntegral (T.length txt)) 0 id
+length =
+    Control.Foldl.Fold (\n txt -> n + fromIntegral (Data.Text.length txt)) 0 id
 {-# INLINABLE length #-}
 
 {-| @(all predicate)@ returns 'True' if all characters satisfy the predicate,
     'False' otherwise
 -}
 all :: (Char -> Bool) -> Fold Text Bool
-all predicate = L.Fold (\b txt -> b && T.all predicate txt) True id
+all predicate =
+    Control.Foldl.Fold (\b txt -> b && Data.Text.all predicate txt) True id
 {-# INLINABLE all #-}
 
 {-| @(any predicate)@ returns 'True' if any character satisfies the predicate,
     'False' otherwise
 -}
 any :: (Char -> Bool) -> Fold Text Bool
-any predicate = L.Fold (\b txt -> b || T.any predicate txt) False id
+any predicate =
+    Control.Foldl.Fold (\b txt -> b || Data.Text.any predicate txt) False id
 {-# INLINABLE any #-}
 
 -- | Computes the maximum character
 maximum :: Fold Text (Maybe Char)
-maximum = L.Fold step Nothing' lazy
+maximum = Control.Foldl.Fold step Nothing' Control.Foldl.Internal.lazy
   where
     step mc txt =
-        if T.null txt
+        if Data.Text.null txt
         then mc
         else Just' (case mc of
-            Nothing' -> T.maximum txt
-            Just' c -> max c (T.maximum txt) )
+            Nothing' -> Data.Text.maximum txt
+            Just' c -> max c (Data.Text.maximum txt) )
 {-# INLINABLE maximum #-}
 
 -- | Computes the minimum character
 minimum :: Fold Text (Maybe Char)
-minimum = L.Fold step Nothing' lazy
+minimum = Control.Foldl.Fold step Nothing' Control.Foldl.Internal.lazy
   where
     step mc txt =
-        if T.null txt
+        if Data.Text.null txt
         then mc
         else Just' (case mc of
-            Nothing' -> T.minimum txt
-            Just' c -> min c (T.minimum txt) )
+            Nothing' -> Data.Text.minimum txt
+            Just' c -> min c (Data.Text.minimum txt) )
 {-# INLINABLE minimum #-}
 
 {-| @(elem c)@ returns 'True' if the text stream has a character equal to @c@,
@@ -136,10 +155,10 @@ notElem c = all (c /=)
     or 'Nothing' if no character satisfies the predicate
 -}
 find :: (Char -> Bool) -> Fold Text (Maybe Char)
-find predicate = L.Fold step Nothing' lazy
+find predicate = Control.Foldl.Fold step Nothing' Control.Foldl.Internal.lazy
   where
     step mc txt = case mc of
-        Nothing' -> strict (T.find predicate txt)
+        Nothing' -> strict (Data.Text.find predicate txt)
         Just' _  -> mc
 {-# INLINABLE find #-}
 
@@ -147,13 +166,13 @@ find predicate = L.Fold step Nothing' lazy
     the stream has an insufficient number of characters
 -}
 index :: Integral n => n -> Fold Text (Maybe Char)
-index i = L.Fold step (Left' (fromIntegral i)) hush
+index i = Control.Foldl.Fold step (Left' (fromIntegral i)) hush
   where
     step x txt = case x of
         Left' remainder ->
-            let len = T.length txt
+            let len = Data.Text.length txt
             in  if remainder < len
-                then Right' (T.index txt remainder)
+                then Right' (Data.Text.index txt remainder)
                 else Left'  (remainder - len)
         _               -> x
 {-# INLINABLE index #-}
@@ -170,22 +189,27 @@ elemIndex c = findIndex (c ==)
     predicate
 -}
 findIndex :: Num n => (Char -> Bool) -> Fold Text (Maybe n)
-findIndex predicate = L.Fold step (Left' 0) hush
+findIndex predicate = Control.Foldl.Fold step (Left' 0) hush
   where
     step x txt = case x of
-        Left' m -> case T.findIndex predicate txt of
-            Nothing -> Left'  (m + fromIntegral (T.length txt))
+        Left' m -> case Data.Text.findIndex predicate txt of
+            Nothing -> Left'  (m + fromIntegral (Data.Text.length txt))
             Just n  -> Right' (m + fromIntegral n)
         _       -> x
 {-# INLINABLE findIndex #-}
 
 -- | @(count c)@ returns the number of times @c@ appears
 count :: Num n => Char -> Fold Text n
-count c = L.Fold step 0 id
+count c = Control.Foldl.Fold step 0 id
   where
-    step n txt = n + fromIntegral (T.count (T.singleton c) txt)
+    step n txt = n + fromIntegral (Data.Text.count (Data.Text.singleton c) txt)
 {-# INLINABLE count #-}
 
+-- | Combine all the strict `Text` chunks to build a lazy `Text`
+lazy :: Fold Text Data.Text.Lazy.Text
+lazy = fmap Data.Text.Lazy.fromChunks Control.Foldl.list
+{-# INLINABLE lazy #-}
+
 {- $reexports
     "Control.Foldl" re-exports the 'Fold' type
 
